@@ -1,4 +1,4 @@
-"""Customize argparse logic for tox (also contains the base options)."""  # noqa: A005
+"""Customize argparse logic for tox (also contains the base options)."""
 
 from __future__ import annotations
 
@@ -11,6 +11,8 @@ from argparse import SUPPRESS, Action, ArgumentDefaultsHelpFormatter, ArgumentEr
 from pathlib import Path
 from typing import TYPE_CHECKING, Any, Callable, Dict, List, Literal, Optional, Sequence, Tuple, Type, TypeVar, cast
 
+from colorama import Fore
+
 from tox.config.loader.str_convert import StrConvert
 from tox.plugin import NAME
 from tox.util.ci import is_ci
@@ -18,6 +20,11 @@ from tox.util.ci import is_ci
 from .env_var import get_env_var
 from .ini import IniConfig
 
+if sys.version_info >= (3, 11):  # pragma: >=3.11 cover
+    from typing import Self
+else:  # pragma: <3.11 cover
+    from typing_extensions import Self
+
 if TYPE_CHECKING:
     from tox.session.state import State
 
@@ -79,7 +86,7 @@ class ArgumentParserWithEnvAndConfig(ArgumentParser):
         res, argv = self.parse_known_args(args, namespace)
         if argv:
             self.error(
-                f'unrecognized arguments: {" ".join(argv)}\n'
+                f"unrecognized arguments: {' '.join(argv)}\n"
                 "hint: if you tried to pass arguments to a command use -- to separate them from tox ones",
             )
         return res
@@ -88,8 +95,8 @@ class ArgumentParserWithEnvAndConfig(ArgumentParser):
 class HelpFormatter(ArgumentDefaultsHelpFormatter):
     """A help formatter that provides the default value and the source it comes from."""
 
-    def __init__(self, prog: str) -> None:
-        super().__init__(prog, max_help_position=30, width=240)
+    def __init__(self, prog: str, **kwargs: Any) -> None:
+        super().__init__(prog, max_help_position=30, width=240, **kwargs)
 
     def _get_help_string(self, action: Action) -> str | None:
         text: str = super()._get_help_string(action) or ""
@@ -123,7 +130,7 @@ class Parsed(Namespace):
     @property
     def is_colored(self) -> bool:
         """:return: flag indicating if the output is colored or not"""
-        return cast(bool, self.colored == "yes")
+        return cast("bool", self.colored == "yes")
 
     exit_and_dump_after: int
 
@@ -205,7 +212,7 @@ class ToxParser(ArgumentParserWithEnvAndConfig):
                     result = None
                 else:
                     try:
-                        result = int(cast(str, values))
+                        result = int(cast("str", values))
                         if result <= 0:
                             msg = "must be greater than zero"
                             raise ValueError(msg)  # noqa: TRY301
@@ -285,11 +292,11 @@ class ToxParser(ArgumentParserWithEnvAndConfig):
         return result
 
     @classmethod
-    def base(cls: type[ToxParserT]) -> ToxParserT:
+    def base(cls) -> Self:
         return cls(add_help=False, root=True)
 
     @classmethod
-    def core(cls: type[ToxParserT]) -> ToxParserT:
+    def core(cls) -> Self:
         return cls(
             prog=NAME,
             formatter_class=HelpFormatter,
@@ -366,6 +373,12 @@ def add_color_flags(parser: ArgumentParser) -> None:
         choices=["yes", "no"],
         help="should output be enriched with colors, default is yes unless TERM=dumb or NO_COLOR is defined.",
     )
+    parser.add_argument(
+        "--stderr-color",
+        default="RED",
+        choices=[*Fore.__dict__.keys()],
+        help="color for stderr output, use RESET for terminal defaults.",
+    )
 
 
 def add_exit_and_dump_after(parser: ArgumentParser) -> None:
