@@ -1,15 +1,18 @@
-"""
-Run tox environments in sequential order.
-"""
+"""Run tox environments in sequential order."""
+
 from __future__ import annotations
 
-from tox.config.cli.parser import ToxParser
+from typing import TYPE_CHECKING
+
 from tox.plugin import impl
-from tox.session.state import State
+from tox.session.env_select import CliEnv, register_env_select_flags
 
-from ...env_select import CliEnv, register_env_select_flags
 from .common import env_run_create_flags, execute
 
+if TYPE_CHECKING:
+    from tox.config.cli.parser import ToxParser
+    from tox.session.state import State
+
 
 @impl
 def tox_add_option(parser: ToxParser) -> None:
